"""Align subtitles using GT transcript and ASR output."""

from subtitle_alchemy.align._index import get_aligned_index
from subtitle_alchemy.align._tl import get_aligned_tl

__all__ = [
    "get_aligned_index",
    "get_aligned_tl",
]



"""Match the ground truth text with the predicted subtitles."""

import numpy as np

from subtitle_alchemy.utils import pinyin

CHR_DEFAULT = "✅"  # Placeholder for missing ground truth or prediction
IDX_DEFAULT = -1  # Placeholder index


def get_aligned_ind_raw(lab: np.ndarray, prd: np.ndarray) -> np.ndarray:
    """Get the raw aligned indices of predictions.

    Args:
        lab (list[str]): The 1D ground truth array of characters.
        prd (list[str]): The 1D predicted array of characters.

    Returns:
        np.ndarray: A list of indices representing the alignment between
          `lab` and `prd`. It has the same length as `lab`, where each entry
          corresponds to either an index in the predicted sequence `prd` or
          `IDX_DEFAULT` if the prediction is missing or incorrect.
          Note that falsely predicted characters are marked with `IDX_DEFAULT`,
          which is not the case for final output.
    """
    n_lab = len(lab)
    n_prd = len(prd)
    seq_prd_idx = list(range(n_prd))

    # DP table to store maximum matches
    dp = [[0] * (n_prd + 1) for _ in range(n_lab + 1)]

    # Fill the DP table
    for i in range(1, n_lab + 1):
        for j in range(1, n_prd + 1):
            if lab[i - 1] == prd[j - 1]:
                dp[i][j] = dp[i - 1][j - 1] + 1
            else:
                dp[i][j] = max(dp[i - 1][j], dp[i][j - 1])

    # Backtracking to determine the aligned prediction and indices
    # seq_chr_aligned = []
    seq_idx_aligned = []
    i_lab, i_prd = n_lab, n_prd
    while i_lab > 0 and i_prd > 0:
        if lab[i_lab - 1] == prd[i_prd - 1]:
            # Case: Match
            # seq_chr_aligned.append(seq_prd[i_prd - 1])
            seq_idx_aligned.append(seq_prd_idx[i_prd - 1])
            i_lab -= 1
            i_prd -= 1
        elif dp[i_lab - 1][i_prd] >= dp[i_lab][i_prd - 1]:
            # Case: missing prediction
            # seq_chr_aligned.append(CHR_DEFAULT)
            seq_idx_aligned.append(IDX_DEFAULT)
            i_lab -= 1
        else:
            # Skip this character in the prediction
            i_prd -= 1

    while i_lab > 0:
        # missing GT
        # seq_chr_aligned.append(CHR_DEFAULT)
        seq_idx_aligned.append(IDX_DEFAULT)
        i_lab -= 1

    # Reverse to get the correct order
    # seq_chr_aligned.reverse()
    seq_idx_aligned.reverse()

    return np.array(seq_idx_aligned, dtype=np.int32)


def locate_mismatch(
    idx: np.ndarray,
    n_prd: int,
) -> tuple[list[list[int]], list[list[int]]]:
    """Find indices of the mis-matched characters in GT and predicted text.

    It depends on the output of the `get_match_idx` function to identify the
    mis-matched characters indices in the ground truth and predicted sequences.

    The function returns two lists:
    - A list of lists, where each sublist contains the indices in the ground
      truth sequence that correspond to missing or incorrectly predicted
      characters.
    - A list of lists, where each sublist contains the indices in the predicted
      sequence that are potential candidates for the missing ground truth
      characters.

    Args:
        idx (np.ndarray): A list of indices representing the alignment between
          `seq_lab` and `seq_prd`. It has the same length as `seq_lab`, where
          each entry corresponds to either an index in the predicted sequence
          `seq_prd` or `IDX_DEFAULT` if the prediction is missing or incorrect.
        n_prd (int): The length of the predicted sequence.

    Returns:
        tuple[list[list[int]], list[list[int]]]: A tuple containing two lists:
        - `sseq_idx_lab_incorrect`: A list where each sublist contains the
          indices in `seq_lab` corresponding to ground truth characters that
          are missing or incorrectly predicted.
        - `sseq_idx_prd_incorrect`: A list where each sublist is a sequence of
          consecutive indices in `seq_prd`, corresponding to one or more ground
          truth characters that are missing or incorrectly predicted.

    Example:
        seq_lab = [..., "轻", "松", "地", "阅", "读", ...]
        seq_prd = [..., "轻", "速", "阿", "的", "阅", "读", ...]
        idx_aligned = [..., 79, -1, -1, 83, 84, ...]

        sseq_idx_lab_incorrect, sseq_idx_prd_incorrect = match_wrong_chrs(
                idx_aligned)

        # Output: [..., [77, 78], ...]; ["松", "地"]
        print(sseq_idx_lab_incorrect)
        # Output: [..., [80, 81, 82], ...]; ["速", "阿", "的"]
        print(sseq_idx_prd_incorrect)
    """
    sseq_idx_lab_incorrect = []
    sseq_idx_prd_incorrect = []
    _idx_lab_last_correct, _idx_prd_last_correct = -1, -1
    _flag_correct = True
    _seq_idx_lab_incorrect = []
    for idx_lab, idx_prd in enumerate(idx):
        if idx_prd == IDX_DEFAULT:
            _seq_idx_lab_incorrect.append(idx_lab)
            _flag_correct = False
        elif _flag_correct is False:
            sseq_idx_prd_incorrect.append(
                list(range(_idx_prd_last_correct + 1, idx_prd))
            )
            sseq_idx_lab_incorrect.append(_seq_idx_lab_incorrect)
            _seq_idx_lab_incorrect = []
            _idx_lab_last_correct, _idx_prd_last_correct = idx_lab, idx_prd
            _flag_correct = True
        else:
            _idx_lab_last_correct, _idx_prd_last_correct = idx_lab, idx_prd
            _flag_correct = True

    # If the last character is incorrect
    if _flag_correct is False:
        sseq_idx_lab_incorrect.append(_seq_idx_lab_incorrect)
        sseq_idx_prd_incorrect.append(
            list(range(_idx_prd_last_correct + 1, n_prd))
        )

    return sseq_idx_lab_incorrect, sseq_idx_prd_incorrect


def _map_lab2prd_miss(
    grp_lab: list[int],
    grp_prd: list[int],
    seq_lab: list[str],
    seq_prd: list[str],
) -> dict[int, int]:
    """Prediction missing characters, helper for `get_uniq_miss_map`."""
    _map = {}
    n_lab = len(grp_lab)
    n_prd = len(grp_prd)

    for i, idx_prd in enumerate(grp_prd):
        idx_lab_maxsim = None  # GT index with max similarity
        sim_max = 0.0
        prd_char = seq_prd[idx_prd]

        # do NOT calculate if out of order
        for j in range(i, i + n_lab - n_prd + 1):
            idx_lab = grp_lab[j]
            lab_char = seq_lab[idx_lab]
            sim_pinyin = pinyin.sim(lab_char, prd_char)

            if sim_pinyin > sim_max:
                sim_max = sim_pinyin
                idx_lab_maxsim = idx_lab

        if idx_lab_maxsim is not None:
            _map[idx_lab_maxsim] = idx_prd

    # loop through the GT
    for idx_lab in grp_lab:
        if idx_lab not in _map:
            _map[idx_lab] = -1

    return _map


def _map_lab2prd_extra(
    grp_lab: list[int],
    grp_prd: list[int],
    seq_lab: list[str],
    seq_prd: list[str],
) -> dict[int, int]:
    """Prediction extra characters, helper for `get_uniq_miss_map`."""
    _map = {}
    n_lab = len(grp_lab)
    n_prd = len(grp_prd)

    for i, idx_lab in enumerate(grp_lab):
        idx_prd_maxsim = None  # prediction index with max similarity
        sim_max = 0.0
        lab_char = seq_lab[idx_lab]

        # do NOT calculate if out of order
        for j in range(i, i + n_prd - n_lab + 1):
            idx_prd = grp_prd[j]
            prd_char = seq_prd[idx_prd]
            sim_pinyin = pinyin.sim(lab_char, prd_char)

            if sim_pinyin > sim_max:
                sim_max = sim_pinyin
                idx_prd_maxsim = idx_prd

        if idx_prd_maxsim is not None:
            _map[idx_lab] = idx_prd_maxsim

    return _map


def get_mismatch_map(
    seq_lab: list[str],
    seq_prd: list[str],
    sseq_idx_lab_incorrect: list[list[int]],
    sseq_idx_prd_incorrect: list[list[int]],
) -> dict[int, int]:
    """Get 1-to-1 mapping of incorrectly predicted GT index to prediction one.

    - The map is from the GT index to the prediction index.
    - There are three cases to consider:
      - Predicted wrong characters: in this case the 1-to-1 mapping is trivial.
      - Missing characters: the prediction is missing one or more characters.
        Note that in this case some GT indices will NOT have corresponding
        prediction indices.
      - Extra characters: the prediction has one or more extra characters.
    - Map incorrectly predicted GT Chinese characters to their closest
      predictions based on phonetic similarity, returning a dictionary from GT
      index to prediction index.

    Args:
        seq_lab (list[str]): The ground truth sequence of characters.
        seq_prd (list[str]): The predicted sequence of characters.
        sseq_idx_lab_incorrect (list[list[int]]): A list where each sublist
            contains the indices in `seq_lab` corresponding to ground truth
            characters that are missing or incorrectly predicted.
        sseq_idx_prd_incorrect (list[list[int]]): A list where each sublist
            contains the indices in `seq_prd` corresponding to one or more
            ground truth characters that are missing or incorrectly predicted.

    Returns:
        dict[int, int]: A dictionary mapping each incorrectly predicted GT
            index to its closest prediction index.
    """
    mapping = {}

    for lab_grp, prd_grp in zip(
        sseq_idx_lab_incorrect, sseq_idx_prd_incorrect, strict=True
    ):
        n_lab = len(lab_grp)
        n_prd = len(prd_grp)
        # 1. Predicted wrong characters
        # This is the simplest case where the 1-to-1 mapping is determined
        if n_prd == n_lab:
            for idx_lab, idx_prd in zip(lab_grp, prd_grp, strict=True):
                mapping[idx_lab] = idx_prd
        # 2. Missing characters
        elif n_prd < n_lab:
            _map_miss = _map_lab2prd_miss(lab_grp, prd_grp, seq_lab, seq_prd)
            mapping.update(_map_miss)

        # 3. Extra characters
        else:
            _map_extra = _map_lab2prd_extra(lab_grp, prd_grp, seq_lab, seq_prd)
            mapping.update(_map_extra)

    return mapping


def get_aligned_index(ar_lab: np.ndarray, ar_prd: np.ndarray) -> np.ndarray:
    """Get aligned indices of predictions.

    Args:
        ar_lab (np.ndarray): The ground truth sequence of characters.
        ar_prd (np.ndarray): The predicted sequence of characters.

    Returns:
        np.ndarray: A list of indices representing the alignment between
          `ar_lab` and `ar_prd`. It has the same length as `ar_lab`, where
          each entry corresponds to either an index in the predicted sequence
          `ar_prd` or `IDX_DEFAULT` if the prediction is missing. Falsely
          predicted character indices are also included.
    """
    _mat_match = {}
    idx_aligned_raw = get_aligned_ind_raw(ar_lab, ar_prd)
    for idx_lab, idx_prd in enumerate(idx_aligned_raw):
        if idx_prd != IDX_DEFAULT:
            _mat_match[idx_lab] = idx_prd
    sseq_idx_lab_mis, sseq_idx_prd_mis = locate_mismatch(
        idx_aligned_raw, len(ar_prd)
    )
    _map_mis = get_mismatch_map(
        ar_lab, ar_prd, sseq_idx_lab_mis, sseq_idx_prd_mis
    )
    _mat_match.update(_map_mis)
    return np.array([v for _, v in sorted(_mat_match.items())], dtype=np.int32)



"""Generate subtitle from saved transcription."""

from pathlib import Path

import click
from loguru import logger
from subtitle_alchemy import merge
from subtitle_alchemy.forger import gen_srt
from subtitle_alchemy.utils import sl


@click.command()
@click.argument("src", type=click.Path(exists=True, file_okay=True))
@click.argument("folder", type=click.Path(exists=False, file_okay=False))
@click.option(
    "--form",
    help="Subtitle format to generate",
    default="srt",
    show_default=True,
    type=str,
)
@click.option(
    "--threshold",
    help="Gap threshold in milliseconds",
    default=500,
    show_default=True,
    type=click.INT,
)
def generate(
    src: str,
    folder: str,
    form: str = "srt",
    threshold: int = 500,
) -> None:
    """Generate subtitle from saved transcription."""
    key, txt, tl, _ = sl.load(Path(src))
    adh = merge.tl2adh(tl, th=threshold)
    instr = merge.adh2instr(adh)
    txt_ = merge.merge_txt(txt, instr)
    tl_ = merge.merge_tl(tl, instr)
    if form == "srt":
        gen_srt(tl_, txt_, Path(folder) / f"{key}.srt")
    else:
        logger.error(f"Unsupported format: {form}")
        raise ValueError()



"""Align ground truth and prediction."""

from pathlib import Path

import click
from loguru import logger
from subtitle_alchemy import merge
from subtitle_alchemy.align import get_aligned_index
from subtitle_alchemy.align import get_aligned_tl
from subtitle_alchemy.forger import gen_srt
from subtitle_alchemy.utils import punc
from subtitle_alchemy.utils import sl


@click.command()
@click.argument("pred", type=click.Path(exists=True, file_okay=True))
@click.argument("truth", type=click.Path(exists=False, file_okay=True))
@click.argument("folder", type=click.Path(exists=False, file_okay=False))
@click.option(
    "--form",
    help="Subtitle format to generate",
    default="srt",
    show_default=True,
    type=str,
)
@click.option(
    "--threshold",
    help="Gap threshold in milliseconds",
    default=500,
    show_default=True,
    type=click.INT,
)
def align(
    pred: str,
    truth: str,
    folder: str,
    form: str = "srt",
    threshold: int = 500,
) -> None:
    """Align predicted transcript with the ground truth speech."""
    p_pred, p_truth, p_folder = Path(pred), Path(truth), Path(folder)
    key, txt_pred, tl_pred, _ = sl.load(p_pred)  # TODO: support punctuation

    with open(p_truth) as f:
        speech = f.read().strip()
    # TODO: support punctuation
    txt_aligned, _ = punc.separate(speech)
    idx_aligned = get_aligned_index(txt_aligned, txt_pred)
    tl_aligned = get_aligned_tl(idx_aligned, tl_pred)

    gap = merge.tl2adh(tl_aligned, th=threshold)
    instr = merge.adh2instr(gap)
    txt_ = merge.merge_txt(txt_aligned, instr)
    tl_ = merge.merge_tl(tl_aligned, instr)
    if form == "srt":
        gen_srt(tl_, txt_, p_folder / f"{key}.srt")
    else:
        logger.error(f"Unsupported format: {form}")
        raise ValueError()



"""Serialization and deserialization of transcript data."""

from pathlib import Path

import numpy as np

_K_TXT = "text"
_K_TL = "timeline"
_K_PUNC = "punc"

__all__ = ["save", "load"]


def save(
    key: str,
    folder: Path,
    txt: np.ndarray,
    tl: np.ndarray,
    punc: np.ndarray,
) -> None:
    """Save transcript as npz file.

    Args:
        key (str): file name
        folder (Path): folder to save the npz file
        txt (np.ndarray): 1D UTF8 encoded array (N, ) of space-separated words
        tl (np.ndarray): 2D int32 timeline array (N, 2) of start-end timestamp
            pairs
        punc (np.ndarray): 1D int32 array (N, ) of punctuations
    """
    np.savez(folder / f"{key}.npz", text=txt, timeline=tl, punc=punc)


def load(path: Path) -> tuple[str, np.ndarray, np.ndarray, np.ndarray]:
    """Load transcript from npz file.

    Args:
        path (Path): npz file path

    Returns:
        tuple[np.ndarray, np.ndarray]: Tuple of words and timeline arrays:
        - key: str file name
        - txt: UTF-8 encoded words array (N, )
        - ts: int32 timeline array (N, 2) i.e. start-end timestamp pairs
        - punc: int32 array (N, ) of punctuations
    """
    with np.load(path) as data:
        return path.stem, data[_K_TXT], data[_K_TL], data[_K_PUNC]



"""Timeline utilities."""

import numpy as np


def ms2hmss(tl: np.ndarray) -> np.ndarray:
    """Convert the milliseconds timeline array to HH:MM:SS.mmm array.

    Args:
        tl (np.ndarray): 2D array (N, 2) of milliseconds

    Returns:
        np.ndarray: 2D array (N, 8) timeline i.e. hours (start), hours (end),
        minutes (start), minutes (end), seconds (start), seconds (end),
        milliseconds (start), milliseconds (end)
    """
    return np.column_stack(
        (
            tl // 3600000,  # hours
            tl // 60000 % 60,  # minutes
            tl // 1000 % 60,  # seconds
            tl % 1000,  # milliseconds
        )
    )


def hmss2ms(hr: int, mn: int, ss: int, ms: int) -> int:
    """Convert the HH:MM:SS.mmm to milliseconds.

    Args:
        hr (int): hours
        mn (int): minutes
        ss (int): seconds
        ms (int): milliseconds

    Returns:
        int: milliseconds
    """
    return hr * 3600000 + mn * 60000 + ss * 1000 + ms



"""STT Service to parse audio files."""

from pathlib import Path

from funasr import AutoModel
from loguru import logger
import numpy as np

__all__ = ["generate"]


def generate(
    model_name: str,
    audio: Path,
    hotword: str = "",
) -> tuple[str, np.ndarray, np.ndarray, np.ndarray]:
    """Generate transcript from audio file.

    The return value is expected as a list with a single dict in it.
    The dict should have the following keys:
    - key: file name
    - txt: space-separated words
    - ts: list of timestamp pair lists

    Args:
        model_name (str): STT Model name
        audio (Path): Path to audio file
        hotword (str): Hot words to detect, separated by space, default ""

    Returns:
        tuple[str, np.ndarray, np.ndarray, np.ndarray]: Tuple of key, words and
            timeline arrays:
        - key: file name
        - words: 1D UTF8 encoded array (N, ) of space-separated words
        - timeline: 2D int32 timeline array (N, 2) of start-end timestamp pairs
        - punctuations: 1D int32 array (N, ) of punctuations
    """
    mdl_stt = AutoModel(
        model=model_name,
        vad_model="fsmn-vad",
        log_level="ERROR",
    )
    mdl_punc = AutoModel(model="ct-punc", model_revision="v2.0.4")
    if hotword == "":
        res = mdl_stt.generate(input=str(audio), batch_size_s=300)
    else:
        res = mdl_stt.generate(
            input=str(audio), batch_size_s=300, hotword=hotword
        )
    if not isinstance(res, list):
        logger.error(f"Expected list, got {type(res)}")
        raise TypeError()

    res = res[0]
    if not isinstance(res, dict):
        logger.error(f"Expected dict in list, got {type(res)}")
        raise TypeError()

    key = res["key"]
    txt = res["text"]
    arr_txt = np.array(txt.split(), dtype="U")
    arr_ts = np.array(res["timestamp"], dtype=np.int32)
    arr_punc = mdl_punc.generate(input=txt)[0]["punc_array"]
    # So characters with no punctuation are 0; this aligns with the utils.punc
    arr_punc = (arr_punc - 1).numpy().astype(np.uint8)
    return key, arr_txt, arr_ts, arr_punc



"""Transcribe audio and save transcription."""

from pathlib import Path

import click
from subtitle_alchemy.parser import stt
from subtitle_alchemy.utils import sl


@click.command()
@click.argument("audio", type=click.Path(exists=True, file_okay=True))
@click.argument("folder", type=click.Path(exists=False, file_okay=False))
@click.option(
    "--model",
    help="STT model, default paraformer-zh",
    default="paraformer-zh",
    show_default=True,
    type=str,
)
@click.option(
    "--hotword",
    help="Hot words to detect, separated by space",
    default="",
    show_default=True,
    type=str,
)
def transcribe(
    audio: str,
    folder: str,
    model: str = "paraformer-zh",
    hotword: str = "",
) -> None:
    """Transcribe audio and save transcription."""
    p_audio = Path(audio)
    p_folder = Path(folder)
    key, txt, tl, punc = stt.generate(model, p_audio, hotword)
    sl.save(key, p_folder, txt, tl, punc)



"""Merge Subtitles.

There are two important types of arrays for merging control:

- Adhesion array: 1D array with only 0s and 1s of shape (N, ) which indicates
  either to merge (0) or not (1) with the next item.
  - The data type is uint8.
  - The last item is always 1.
  - It can be obtained easily with for example the timeline array or the
    punctuation array, but not straight-forward enough to do the merging, so we
    need to convert it to the instruction array.
- Instruction array: 2D array of shape (M, 2) which contains the start and end
  indices to merge. It can be obtained by the gap array.
"""

import numpy as np

_K_ADHESION_V_M = 0  # adhesion value for merging
_K_ADHESION_V_S = 1  # adhesion value for splitting

__all__ = ["tl2adh", "adh2instr", "merge_tl", "merge_txt"]


def tl2adh(tl: np.ndarray, th: int = 500) -> np.ndarray:
    """Compute the adhesion between each timestamp pair.

    Args:
        tl (np.ndarray): 2D array (N, 2) of start and end timelines
        th (int): The threshold in milliseconds to determine the gap, default
            is 500ms

    Returns:
        np.ndarray: 1D array (N, ) adhesion array for computing merge
            instructions. One `1` must be appended to the end of the array.
    """
    gap_ms = tl[1:, 0] - tl[:-1, 1]
    gap = np.where(gap_ms <= th, _K_ADHESION_V_M, _K_ADHESION_V_S)
    return np.append(gap, _K_ADHESION_V_S).astype(np.uint8)


def adh2instr(adh: np.ndarray) -> np.ndarray:
    """Get merge instructions from a (N, ) adhesion array.

    Args:
        adh (np.ndarray): 1D array (N, ) of 0s and 1s

    Returns:
        np.ndarray: (M, 2) merge instruction array which contains the start and
            end indices to merge
    """
    # Initialize list to store result pairs
    result = []
    lo = 0

    # Iterate through the array
    for idx, val in enumerate(adh):
        # If the 1st value is 1, add a pair of 0s
        if val == _K_ADHESION_V_S and idx == 0:
            result.append([0, 0])
            lo = idx + 1
        # If the current value and the previous value are 1, add a pair
        elif (
            val == _K_ADHESION_V_S
            and idx > 0
            and adh[idx - 1] == _K_ADHESION_V_S
        ):
            result.append([idx, idx])
            lo = idx + 1
        # Consecutive 0s ended
        elif val == _K_ADHESION_V_S and idx > 0 and adh[idx - 1] == 0:
            result.append([lo, idx])
            lo = idx + 1
        # Ignore consecutive 0s
        else:
            continue

    return np.array(result, dtype=np.int32)


def merge_tl(tl: np.ndarray, instr: np.ndarray) -> np.ndarray:
    """Merge the timeline array based on a merge instruction array.

    Args:
        tl (np.ndarray): 2D array (N, 2) of start and end in milliseconds
        instr (np.ndarray): 2D array (M, 2) merge instruction array which
          contains the start and end indices to merge

    Returns:
        np.ndarray: 2D array (M, 2) of merged start and end timestamps
    """
    return np.column_stack((tl[instr[:, 0], 0], tl[instr[:, 1], 1]))


def merge_txt(txt: np.ndarray, instr: np.ndarray) -> np.ndarray:
    """Merge the text array based on a merge instruction array.

    Args:
        txt (np.ndarray): 1D array of UTF-8 encoded strings
        instr (np.ndarray): 2D array (M, 2) merge instruction array which
          contains the start and end indices to merge

    Returns:
        np.ndarray: 1D array (M, ) of merged strings
    """
    return np.array(["".join(txt[s : e + 1]) for s, e in instr])



"""Parse SRT files."""

import re

import numpy as np

from subtitle_alchemy.utils import timeline

N_BLOCK_LINES = 3


def generate(srt_string: str) -> tuple[np.ndarray, np.ndarray]:
    """Parse SRT string into arrays of lines and timestamps.

    Args:
        srt_string (str): Input SRT string.

    Returns:
        tuple[np.ndarray, np.ndarray]: Tuple of
        - lines: Array of subtitle text lines (N,)
        - timestamps: Array of start/end timestamps in seconds (N, 2)
    """
    # Split into subtitle blocks
    blocks = re.split(r"\n\n+", srt_string.strip())

    lines = []
    timestamps = []

    for block in blocks:
        if not block.strip():
            continue

        # Split block into lines
        block_lines = block.split("\n")
        if len(block_lines) < N_BLOCK_LINES:
            continue

        # Parse timestamp line (second line)
        tl_match = re.match(
            r"(\d{2}):(\d{2}):(\d{2}),(\d{3}) --> "
            r"(\d{2}):(\d{2}):(\d{2}),(\d{3})",
            block_lines[1],
        )
        if not tl_match:
            continue

        # Convert timestamp to milliseconds
        times = [int(x) for x in tl_match.groups()]
        ms_start = timeline.hmss2ms(*times[:4])
        ms_end = timeline.hmss2ms(*times[4:])

        # Join remaining lines as subtitle text
        subtitle_text = " ".join(block_lines[2:])

        lines.append(subtitle_text)
        timestamps.append([ms_start, ms_end])

    return np.array(lines, dtype="U"), np.array(timestamps, dtype=np.int64)



"""Timeline utility functions."""

from pathlib import Path

import numpy as np

from subtitle_alchemy.utils import timeline


def _hmss2srttl(tl: np.ndarray) -> np.ndarray:
    """Format the HH:MM:SS.mmm array to SRT timeline string.

    Args:
        tl (np.ndarray): 2D array (N, 8) of hours, minutes, seconds, and ms

    Returns:
        np.ndarray: 1D array (N,) of SRT timeline strings
        - 00:00:07,000 --> 00:00:10,500
    """
    return np.array(
        [
            f"{hmss[0]:02d}:{hmss[2]:02d}:{hmss[4]:02d},{hmss[6]:03d} --> "
            f"{hmss[1]:02d}:{hmss[3]:02d}:{hmss[5]:02d},{hmss[7]:03d}"
            for hmss in tl
        ],
        dtype="U29",
    )


def gen_srt(ms: np.ndarray, text: np.ndarray, path: Path) -> None:
    """Generate an SRT file from the timeline and text arrays.

    Args:
        ms (np.ndarray): 2D array (N, 2) of milliseconds
        text (np.ndarray): 1D array (N,) of text
        path (Path): Path to save the SRT file
    """
    hmss = timeline.ms2hmss(ms)
    tl_str = _hmss2srttl(hmss)

    with open(path, "w") as f:
        for i, (hmss, txt) in enumerate(zip(tl_str, text, strict=True)):
            f.write(f"{i+1}\n")
            f.write(f"{hmss}\n")
            f.write(f"{txt}\n")
            f.write("\n")



"""Align timeline with alignment index."""

import numpy as np


def acc_default(idx: np.ndarray) -> np.ndarray:
    """Get an array with the count of accumulated default values.

    Example:
        input:  [0, -1, 1, 2, 3, 4, 5, -1, -1, -1, 7, 8, 9]
        output: [0,  1, 0, 0, 0, 0, 0,  1,  2,  3, 0, 0, 0]

    Args:
        idx (np.ndarray): the alignment index with sparse -1 values.

    Returns:
        np.ndarray: the alignment index with -1 values imputed with its
            accumulated count.
    """
    # Initialize the output array with zeros
    ret = np.zeros_like(idx, dtype=np.int32)

    # Variable to track the count of consecutive `-1`s
    count = 0

    # Iterate over the array
    for i in range(idx.size):
        if idx[i] == -1:
            count += 1
            ret[i] = count  # Set the output to the count
        else:
            count = 0  # Reset count when it's not a `-1`

    return ret


def cum_default(acc: np.ndarray) -> np.ndarray:
    """Get an array with the total count of consecutive default values.

    Note that the input should be the output of the `acc_default` function.

    """
    ret = np.zeros_like(acc)

    # Track the current maximum in the increasing series
    max_cur = 0

    # Iterate through the array backwards
    for i in range(acc.size - 1, -1, -1):
        if acc[i] == 0:
            # Reset the current max when we encounter a 0
            max_cur = 0
        else:
            # Update the current max (it keeps the largest value in the series)
            max_cur = max(max_cur, acc[i])
            # Store the current max in the result array
            ret[i] = max_cur

    return ret


def impute_pre(idx: np.ndarray) -> np.ndarray:
    """Impute negative values with the its predecessor.

    Args:
        idx (np.ndarray): the alignment index with sparse -1 values.

    Returns:
        np.ndarray: the alignment index with -1 values imputed with its
            predecessor.
    """
    # copy the input array
    idx = idx.copy()

    # Get the indices where the elements are default values (-1)
    indices_default = np.where(idx == -1)[0]

    # Iterate over the indices and replace -1 with the previous element
    for i in indices_default:
        if i > 0:  # Ensure there is a predecessor
            idx[i] = idx[i - 1]

    return idx


def impute_suc(idx: np.ndarray) -> np.ndarray:
    """Impute negative values with the its successors.

    Args:
        idx (np.ndarray): the alignment index with sparse -1 values.

    Returns:
        np.ndarray: the alignment index with -1 values imputed with its
            successors.
    """
    # copy the input array
    arr = idx.copy()

    # Get the indices where the elements are -1
    indices_default = np.where(arr == -1)[0]

    # Iterate backwards over the indices and replace -1 with the next element
    for i in reversed(indices_default):
        if i < len(arr) - 1:  # Ensure there is a successor
            arr[i] = arr[i + 1]

    return arr


def get_miss_interval(
    acc: np.ndarray,
    cum: np.ndarray,
    pre: np.ndarray,
    suc: np.ndarray,
    tl: np.ndarray,
) -> np.ndarray:
    """Get the missing timeline intervals, and other necessary information."""
    # Get the indices of default values
    indices_diff = np.where(cum >= 1)[0]
    n_diff = indices_diff.size
    interval_diff = np.zeros((n_diff, 5), dtype=np.int32)
    interval_diff[:, 0] = acc[indices_diff]
    interval_diff[:, 1] = cum[indices_diff]
    interval_diff[:, 4] = indices_diff
    for k, v in enumerate(indices_diff):
        interval_diff[k, 2] = tl[pre[v], 1]
        interval_diff[k, 3] = tl[suc[v], 0]
    return interval_diff


def _get_miss_tl_start(idx: int, cnt: int, start: int, end: int) -> int:
    """Get the start time of the missing timeline.

    Note: need to be vectorized.
    """
    # only one missing, or the first missing
    if cnt == 1 or idx == 1:
        return start
    intvl = (end - start) // cnt
    return start + (idx - 1) * intvl


def _get_miss_tl_end(idx: int, cnt: int, start: int, end: int) -> int:
    """Get the end time of the missing timeline.

    Note: need to be vectorized.
    """
    # only one missing, or the last missing
    if cnt in (1, idx):
        return end
    intvl = (end - start) // cnt
    return start + idx * intvl


def get_miss_tl(idx_aligned: np.ndarray, tl: np.ndarray) -> np.ndarray:
    """Interpolate the missing timeline.

    To interpolate the missing timeline, we need to know:
    - the earliest possible start time (end of the previous matched timeline)
    - the latest possible end time (start of the next matched timeline)
    - if it belongs to a consecutive missing timeline, we also need to know:
      - total number of missing timelines
      - index of this missing item in the consecutive missing series
    We approach this using the most naive way: separate the longest possible
    interval evenly before assignment.

    Args:
        idx_aligned (np.ndarray): the 1D alignment index array (N, ) with
            sparse default values.
        tl (np.ndarray): the 2D timeline array (N, 2) with start and end time
            pairs.

    Returns:
        np.ndarray: 2D array (N, 3) with the missing timelines:
        - (N, 0): the index of `N`, indicating the missing location.
        - (N, 1): the start time of the missing timeline.
        - (N, 2): the end time of the missing timeline.
    """
    # TODO: move acc_default and cum_default to `_idx.py`
    acc = acc_default(idx_aligned)
    cum = cum_default(acc)

    idx_pre = impute_pre(idx_aligned)
    idx_suc = impute_suc(idx_aligned)
    # TODO: if the missing interval is too long, it is possible that the
    #   the content was discarded by the speaker. Consider to add a threshold,
    #   and and a function to align the text.
    intvl = get_miss_interval(acc, cum, idx_pre, idx_suc, tl)
    tl_start = np.vectorize(_get_miss_tl_start, otypes=[np.int32])(
        intvl[:, 0], intvl[:, 1], intvl[:, 2], intvl[:, 3]
    )
    tl_end = np.vectorize(_get_miss_tl_end, otypes=[np.int32])(
        intvl[:, 0], intvl[:, 1], intvl[:, 2], intvl[:, 3]
    )
    return np.column_stack((intvl[:, 4], tl_start, tl_end))


def get_match_tl(idx_aligned: np.ndarray, tl: np.ndarray) -> np.ndarray:
    """Get the matched timeline.

    Args:
        idx_aligned (np.ndarray): the 1D alignment index array (N, ) with
            sparse default values.
        tl (np.ndarray): the 2D timeline array (N, 2) with start and end time
            pairs.

    Returns:
        np.ndarray: 2D array (N, 3) with the matched timelines:
        - (N, 0): the index of `N`, indicating the matched location.
        - (N, 1): the start time of the matched timeline.
        - (N, 2): the end time of the matched timeline.
    """
    idx = np.where(idx_aligned != -1)[0]
    tl_ = tl[idx_aligned[idx]]
    return np.column_stack((idx, tl_))


def get_aligned_tl(idx_aligned: np.ndarray, tl: np.ndarray) -> np.ndarray:
    """Get the aligned timeline.

    Args:
        idx_aligned (np.ndarray): the 1D alignment index array (N, ) with
            sparse default values.
        tl (np.ndarray): the 2D timeline array (N, 2) with start and end time
            pairs.

    Returns:
        np.ndarray: 2D array (N, 3) with the aligned timelines:
        - (N, 0): the index of `N`, indicating the aligned location.
        - (N, 1): the start time of the aligned timeline.
        - (N, 2): the end time of the aligned timeline.
    """
    tl_miss = get_miss_tl(idx_aligned, tl)
    tl_match = get_match_tl(idx_aligned, tl)
    tl = np.concatenate((tl_miss, tl_match), axis=0)
    return tl[tl[:, 0].argsort()][:, 1:].astype(np.int32)



"""All CLI commands here."""

import click
from subtitle_alchemy.cli.align import align
from subtitle_alchemy.cli.generate import generate
from subtitle_alchemy.cli.squash import squash
from subtitle_alchemy.cli.transcribe import transcribe


@click.group()
def cli() -> None:
    """Subalch CLI."""


cli.add_command(align)
cli.add_command(generate)
cli.add_command(squash)
cli.add_command(transcribe)



"""Separate text (Chinese characters and ASCII words) with punctuation."""
# ruff: noqa: RUF001

import re

import numpy as np

# Regular expression to match Chinese characters
PATTERN_ZH = r"[\u4e00-\u9fff]"

# Regular expression to match ASCII words (sequences of A-Z, a-z, 0-9)
PATTERN_ASCII = r"[a-zA-Z0-9]+"

# Combined word patterns to match Chinese characters or ASCII words
PATTERN_WORD = f"({PATTERN_ZH})|({PATTERN_ASCII})"

_K_NONE = 0  # No punctuation
# Dictionary to map punctuation marks to integers
MAP_PUNC2INT = {
    "，": 1,  # Chinese comma
    "。": 2,  # Chinese period
    "！": 3,  # Chinese exclamation mark
    "？": 4,  # Chinese question mark
    "、": 5,  # Chinese enumeration mark
    "；": 6,  # Chinese semicolon
    "：": 7,  # Chinese colon
    "“": 8,  # Chinese left double quotation mark
    "”": 9,  # Chinese right double quotation mark
    "‘": 10,  # Chinese left single quotation mark
    "’": 11,  # Chinese right single quotation mark
    "「": 12,  # Chinese left corner bracket
    "」": 13,  # Chinese right corner bracket
    "『": 14,  # Chinese left white corner bracket
    "』": 15,  # Chinese right white corner bracket
    "《": 16,  # Chinese left double angle bracket
    "》": 17,  # Chinese right double angle bracket
    "（": 18,  # Chinese left parenthesis
    "）": 19,  # Chinese right parenthesis
    "【": 20,  # Chinese left square bracket
    "】": 21,  # Chinese right square bracket
    "〈": 22,  # Chinese left angle bracket
    "〉": 23,  # Chinese right angle bracket
    ",": 24,  # English comma
    ".": 25,  # English period
    "!": 26,  # English exclamation mark
    "?": 27,  # English question mark
}
MAP_INT2PUNC = {v: k for k, v in MAP_PUNC2INT.items()}


def separate(text: str) -> tuple[np.ndarray, np.ndarray]:
    """Separate text (Chinese characters and ASCII words) with punctuation.

    TODO | NOTE: If there are multiple punctuation marks consecutively
    (e.g., !?!), the script only identifies the first one.

    Args:
        text (str): The input string containing Chinese characters, ASCII words
            and punctuation marks.

    Returns:
        tuple[np.ndarray, np.ndarray]: A tuple containing two numpy arrays:
        - A UTF-8 encoded 1D array of strings where each element is a Chinese
          character or ASCII word.
        - A 1D uint16 array of integers where each element:
          - 0 if there's no punctuation after the corresponding element
          - a positive integer if there is punctuation (1: comma, 2: period,
            etc.).
    """
    len_text = len(text)
    # Find all matches in the text (including the index positions)
    words = re.finditer(PATTERN_WORD, text)

    # Lists to store results and punctuation info
    seq_txt = []
    seq_mark = []

    # Iterate over all matches to extract the words/characters and check for
    # punctuation
    for word in words:
        # Append the matched text (either a Chinese character or ASCII word)
        seq_txt.append(word.group())

        # Check the character right after the current match
        match_end = word.end()
        if match_end < len_text and text[match_end] in MAP_PUNC2INT:
            # If the next character is punctuation, append its mapped value
            seq_mark.append(MAP_PUNC2INT[text[match_end]])
        else:
            # Otherwise no punctuation
            seq_mark.append(_K_NONE)

    return np.array(seq_txt, dtype="U"), np.array(seq_mark, dtype=np.uint16)


def restore(text: np.ndarray, punc: np.ndarray) -> np.ndarray:
    """Restore punctuation marks to the text.

    Args:
        text (np.ndarray): A 1D UTF-8 encoded array (N, ) of strings where each
            element is a Chinese character or an ASCII word.
        punc (np.ndarray): A 1D uint16 array of integers where each element:
            - 0 if there's no punctuation after the corresponding element
            - a positive integer if there is punctuation (1: comma, 2: period,
              etc.).

    Returns:
        np.ndarray: A 1D UTF-8 encoded array (N, ) of strings where each
        element is a Chinese character or an ASCII word, which may be followed
        by a punctuation mark.
    """
    # Initialize an empty list to store the restored text
    restored = []

    # Iterate over the text and punctuation arrays
    for _, (word, mark) in enumerate(zip(text, punc, strict=True)):
        if mark == _K_NONE:
            # If there's no punctuation, append the word/character
            restored.append(word)
        else:
            # Otherwise append punctuation mark to the word and then append
            # to the list
            restored.append(word + MAP_INT2PUNC[mark])

    return np.array(restored, dtype="U")
